/*-------------------------------------------------------------------------------

  File        : dtmri_view.cpp

  Description : A viewer of Diffusion-Tensor MRI volumes (medical imaging).

  Copyright  : David Tschumperle - http://www.greyc.ensicaen.fr/~dtschump/
  
  This software is governed by the CeCILL-C license under French law and
  abiding by the rules of distribution of free software.  You can  use, 
  modify and/ or redistribute the software under the terms of the CeCILL-C
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info". 
  
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability. 
  
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or 
  data to be ensured and,  more generally, to use and operate it in the 
  same conditions as regards security. 
  
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-C license and that you accept its terms.
  
  ------------------------------------------------------------------------------*/

#include "../CImg.h"
using namespace cimg_library;
// The undef below is necessary when using a non-standard compiler.
#ifdef cimg_use_visualcpp6
#define std
#endif

// Compute fractional anisotropy (FA) of a tensor
//-------------------------------------------
template<typename T> float get_FA(const T& val1, const T& val2, const T& val3) {
  const float 
    l1 = val1>0?val1:0, l2 = val2>0?val2:0, l3 = val3>0?val3:0,
    lm = (l1+l2+l3)/3,
    tr2 = 2*( l1*l1 + l2*l2 + l3*l3 ),
    ll1 = l1-lm,
    ll2 = l2-lm,
    ll3 = l3-lm;
  if (tr2>0) return (float)std::sqrt( 3*(ll1*ll1 + ll2*ll2 + ll3*ll3)/tr2 );
  return 0;
}

// Insert an ellipsoid in a CImg 3D scene
//----------------------------------------
template<typename t,typename tp,typename tf,typename tc>
void insert_ellipsoid(const CImg<t>& tensor,const float X,const float Y,const float Z,const float tfact,
		      const float vx, const float vy, const float vz,
		      CImgList<tp>& points, CImgList<tf>& faces, CImgList<tc>& colors,
		      const unsigned int res1 = 20, const unsigned int res2 = 20) {
  
  // Compute eigen elements
  const float l1 = tensor[0], l2 = tensor[1], l3 = tensor[2], fa = get_FA(l1,l2,l3);

  CImg<> vec = CImg<>::matrix(tensor[3],tensor[6],tensor[9],
			      tensor[4],tensor[7],tensor[10],
			      tensor[5],tensor[8],tensor[11]);
  const int
    r = cimg::min(30+(int)cimg::abs(255*fa*tensor[3]),255),
    g = cimg::min(30+(int)cimg::abs(255*fa*tensor[4]),255),
    b = cimg::min(30+(int)cimg::abs(255*fa*tensor[5]),255);
  
  // Define mesh points
  const unsigned int N0 = points.size;
  for (unsigned int v=1; v<res2; v++)
    for (unsigned int u=0; u<res1; u++) {
      const float 
	alpha = (float)(u*2*cimg::PI/res1),
	beta = (float)(-cimg::PI/2 + v*cimg::PI/res2),
	x = (float)(tfact*l1*std::cos(beta)*std::cos(alpha)),
	y = (float)(tfact*l2*std::cos(beta)*std::sin(alpha)),
	z = (float)(tfact*l3*std::sin(beta));
      points.insert((CImg<tp>::vector(X,Y,Z)+vec*CImg<tp>::vector(x,y,z)).mul(CImg<tp>::vector(vx,vy,vz)));
    }
  const unsigned int N1 = points.size;
  points.insert((CImg<tp>::vector(X,Y,Z)+vec*CImg<tp>::vector(0,0,-l3*tfact)));
  points.insert((CImg<tp>::vector(X,Y,Z)+vec*CImg<tp>::vector(0,0,l3*tfact)));
  points[points.size-2](0)*=vx; points[points.size-2](1)*=vy;  points[points.size-2](2)*=vz;
  points[points.size-1](0)*=vx; points[points.size-1](1)*=vy;  points[points.size-1](2)*=vz;
  
  // Define mesh triangles  
  for (unsigned int vv=0; vv<res2-2; vv++)
    for (unsigned int uu=0; uu<res1; uu++) {
      const int nv = (vv+1)%(res2-1), nu = (uu+1)%res1;
      faces.insert(CImg<tf>::vector(N0+res1*vv+nu,N0+res1*nv+uu,N0+res1*vv+uu));
      faces.insert(CImg<tf>::vector(N0+res1*vv+nu,N0+res1*nv+nu,N0+res1*nv+uu));
      colors.insert(CImg<tc>::vector(r,g,b));
      colors.insert(CImg<tc>::vector(r,g,b));
    }
  for (unsigned int uu=0; uu<res1; uu++) {
    const int nu = (uu+1)%res1;
    faces.insert(CImg<tf>::vector(N0+nu,N0+uu,N1));
    faces.insert(CImg<tf>::vector(N0+res1*(res2-2)+nu, N1+1,N0+res1*(res2-2)+uu));
    colors.insert(CImg<tc>::vector(r,g,b));
    colors.insert(CImg<tc>::vector(r,g,b));
  }
}

// Insert a fiber in a CImg 3D scene
//-----------------------------------
template<typename T,typename te,typename tp, typename tf, typename tc>
void insert_fiber(const CImg<T>& fiber, const CImg<te>& eigen, const CImg<tc>& texture,
		  const int xm, const int ym, const int zm,
		  const float vx, const float vy, const float vz,
		  CImgList<tp>& points, CImgList<tf>& primitives, CImgList<tc>& colors) {
  const int N0 = points.size;
  float x0 = fiber(0,0), y0 = fiber(0,1), z0 = fiber(0,2), fa0 = eigen.linear_pix3d(x0,y0,z0,12);
  points.insert(CImg<>::vector(vx*(x0-xm),vy*(y0-ym),vz*(z0-zm)));
  for (unsigned int l=1; l<fiber.width; l++) {
    float x1 = fiber(l,0), y1 = fiber(l,1), z1 = fiber(l,2), fa1 = eigen.linear_pix3d(x1,y1,z1,12);
    points.insert(CImg<tp>::vector(vx*(x1-xm),vy*(y1-ym),vz*(z1-zm)));
    primitives.insert(CImg<tf>::vector(N0+l-1,N0+l,(int)(fa0*255),0,(int)(fa1*255),0));
    colors.insert(texture.get_shared());
    x0=x1; y0=y1; z0=z1; fa0=fa1;
  }  
}

// Compute fiber tracking using 4th-order Runge Kutta integration
//-----------------------------------------------------------------
template<typename T>
CImg<> get_fibertrack(const CImg<T>& eigen,
		      const float X0, const float Y0, const float Z0, const float lmax=100,
		      const float dl = 0.1f, const float FAmin=0.7f, const float cmin=0.5f) {
  
  CImgList<> resf;
  float normU=0,normpU=0;

  // Forward tracking
  float l = 0, X = X0, Y = Y0, Z = Z0;
  T
    pu = eigen.linear_pix3d(X0,Y0,Z0,3),
    pv = eigen.linear_pix3d(X0,Y0,Z0,4),
    pw = eigen.linear_pix3d(X0,Y0,Z0,5);
  normpU = (float)std::sqrt(pu*pu+pv*pv+pw*pw);
  bool stopflag = false;
  
  while (!stopflag) {
    if (X<0 || X>eigen.dimx()-1 || Y<0 || Y>eigen.dimy()-1 || Z<0 || Z>eigen.dimz()-1 ||
	eigen((int)X,(int)Y,(int)Z,12)<FAmin || l>lmax) stopflag = true;
    else {
      resf.insert(CImg<>::vector(X,Y,Z));
      T 
	u0 = 0.5f*dl*eigen.linear_pix3d(X,Y,Z,3),
	v0 = 0.5f*dl*eigen.linear_pix3d(X,Y,Z,4),
	w0 = 0.5f*dl*eigen.linear_pix3d(X,Y,Z,5);
      if (u0*pu+v0*pv+w0*pw<0) { u0=-u0; v0=-v0; w0=-w0; }
      T
	u1 = 0.5f*dl*eigen.linear_pix3d(X+u0,Y+v0,Z+w0,3),
	v1 = 0.5f*dl*eigen.linear_pix3d(X+u0,Y+v0,Z+w0,4),
	w1 = 0.5f*dl*eigen.linear_pix3d(X+u0,Y+v0,Z+w0,5);
      if (u1*pu+v1*pv+w1*pw<0) { u1=-u1; v1=-v1; w1=-w1; }
      T 
	u2 = 0.5f*dl*eigen.linear_pix3d(X+u1,Y+v1,Z+w1,3),
	v2 = 0.5f*dl*eigen.linear_pix3d(X+u1,Y+v1,Z+w1,4),
	w2 = 0.5f*dl*eigen.linear_pix3d(X+u1,Y+v1,Z+w1,5);
      if (u2*pu+v2*pv+w2*pw<0) { u2=-u2; v2=-v2; w2=-w2; }
      T
	u3 = 0.5f*dl*eigen.linear_pix3d(X+u2,Y+v2,Z+w2,3),
	v3 = 0.5f*dl*eigen.linear_pix3d(X+u2,Y+v2,Z+w2,4),
	w3 = 0.5f*dl*eigen.linear_pix3d(X+u2,Y+v2,Z+w2,5);
      if (u3*pu+v3*pv+w3*pw<0) { u3=-u3; v3=-v3; w3=-w3; }
      T 
	u = u0/3 + 2*u1/3 + 2*u2/3 + u3/3,
	v = v0/3 + 2*v1/3 + 2*v2/3 + v3/3,
	w = w0/3 + 2*w1/3 + 2*w2/3 + w3/3;
      
      normU = (float)std::sqrt(u*u+v*v+w*w);
      const float scal = (u*pu+v*pv+w*pw)/(normU*normpU);
      if (scal<cmin) stopflag=true;

      X+=(pu=u);
      Y+=(pv=v);
      Z+=(pw=w);
      normpU=normU;
      l+=dl;
    }
  }

  // Backward tracking
  l = 0; X = X0; Y = Y0; Z = Z0;
  pu = eigen.linear_pix3d(X0,Y0,Z0,3);
  pv = eigen.linear_pix3d(X0,Y0,Z0,4);
  pw = eigen.linear_pix3d(X0,Y0,Z0,5);
  normpU = (float)std::sqrt(pu*pu+pv*pv+pw*pw);
  stopflag = false;

  while (!stopflag) {
    if (X<0 || X>eigen.dimx()-1 || Y<0 || Y>eigen.dimy()-1 || Z<0 || Z>eigen.dimz()-1 ||
	eigen((int)X,(int)Y,(int)Z,12)<FAmin || l>lmax) stopflag = true;
    else {
      resf.insert(CImg<>::vector(X,Y,Z),0);
      T 
	u0 = 0.5f*dl*eigen.linear_pix3d(X,Y,Z,3),
	v0 = 0.5f*dl*eigen.linear_pix3d(X,Y,Z,4),
	w0 = 0.5f*dl*eigen.linear_pix3d(X,Y,Z,5);
      if (u0*pu+v0*pv+w0*pw<0) { u0=-u0; v0=-v0; w0=-w0; }
      T
	u1 = 0.5f*dl*eigen.linear_pix3d(X+u0,Y+v0,Z+w0,3),
	v1 = 0.5f*dl*eigen.linear_pix3d(X+u0,Y+v0,Z+w0,4),
	w1 = 0.5f*dl*eigen.linear_pix3d(X+u0,Y+v0,Z+w0,5);
      if (u1*pu+v1*pv+w1*pw<0) { u1=-u1; v1=-v1; w1=-w1; }
      T 
	u2 = 0.5f*dl*eigen.linear_pix3d(X+u1,Y+v1,Z+w1,3),
	v2 = 0.5f*dl*eigen.linear_pix3d(X+u1,Y+v1,Z+w1,4),
	w2 = 0.5f*dl*eigen.linear_pix3d(X+u1,Y+v1,Z+w1,5);
      if (u2*pu+v2*pv+w2*pw<0) { u2=-u2; v2=-v2; w2=-w2; }
      T
	u3 = 0.5f*dl*eigen.linear_pix3d(X+u2,Y+v2,Z+w2,3),
	v3 = 0.5f*dl*eigen.linear_pix3d(X+u2,Y+v2,Z+w2,4),
	w3 = 0.5f*dl*eigen.linear_pix3d(X+u2,Y+v2,Z+w2,5);
      if (u3*pu+v3*pv+w3*pw<0) { u3=-u3; v3=-v3; w3=-w3; }
      T 
	u = u0/3 + 2*u1/3 + 2*u2/3 + u3/3,
	v = v0/3 + 2*v1/3 + 2*v2/3 + v3/3,
	w = w0/3 + 2*w1/3 + 2*w2/3 + w3/3;
      
      normU = (float)std::sqrt(u*u+v*v+w*w);
      const float scal = (u*pu+v*pv+w*pw)/(normU*normpU);
      if (scal<cmin) stopflag=true;

      X-=(pu=u);
      Y-=(pv=v);
      Z-=(pw=w);
      normpU=normU;
      l+=dl;
    }
  }

  return resf.get_append('x');  
}

// Main procedure
//----------------
int main(int argc,char **argv) {

  // Read and init data
  //--------------------
  cimg_usage("A viewer of Diffusion-Tensor MRI volumes.");
  const char *file_i   = cimg_option("-i",(char*)0,"Input : Filename of tensor field (volume wxhxdx6)");
  const char* vsize    = cimg_option("-vsize","1x1x1","Input : Voxel aspect");
  const bool normalize = cimg_option("-normalize",true,"Input : Enable tensor normalization");
  const char *file_f   = cimg_option("-f",(char*)0,"Input : Input fibers\n");
  const float dl       = cimg_option("-dl",0.5f,"Fiber computation : Integration step");
  const float famin    = cimg_option("-famin",0.3f,"Fiber computation : Fractional Anisotropy threshold");
  const float cmin     = cimg_option("-cmin",0.2f,"Fiber computation : Curvature threshold");
  const float lmin     = cimg_option("-lmin",10.0f,"Fiber computation : Minimum length\n"); 
  const float lmax     = cimg_option("-lmax",1000.0f,"Fiber computation : Maximum length\n"); 
  const float tfact    = cimg_option("-tfact",1.2f,"Display : Tensor size factor");
  
  CImg<> tensors;
  if (file_i) {
    std::fprintf(stderr,"\n- Loading tensors '%s'",cimg::basename(file_i));
    tensors.load(file_i);
  } else {
    // Create a synthetic tensor field here
    std::fprintf(stderr,"\n- No input files : Creating a synthetic tensor field");
    tensors.assign(32,32,32,6);
    const CImg<> Id = CImg<>::diagonal(0.3f,0.3f,0.3f);
    cimg_mapXYZ(tensors,x,y,z) {
      const float
	u = x-tensors.dimx()/2.0f,
	v = y-tensors.dimy()/2.0f,
	w = z-tensors.dimz()/2.0f,
	norm = (float)std::sqrt(1e-5f+u*u+v*v+w*w),
	nu = u/norm, nv = v/norm, nw = w/norm;
      const CImg<>
	dir1 = CImg<>::vector(nu,nv,nw),
	dir2 = CImg<>::vector(-nv,nu,nw),
	dir3 = CImg<>::vector(nw*(nv-nu),-nw*(nu+nv),nu*nu+nv*nv);
      tensors.set_tensor_at(2.0f*dir1*dir1.get_transpose() +
			    1.0f*dir2*dir2.get_transpose() +
			    0.7f*dir3*dir3.get_transpose(),
			    x,y,z);
    }
  }
  float voxw=1,voxh=1,voxd=1;
  std::sscanf(vsize,"%f%*c%f%*c%f",&voxw,&voxh,&voxd);
  
  std::fprintf(stderr," : %ux%ux%u image, voxsize=%gx%gx%g.",
	       tensors.width,tensors.height,tensors.depth,
	       voxw,voxh,voxd);


  CImgList<> fibers;
  if (file_f) {
    std::fprintf(stderr,"\n- Loading fibers '%s'.",cimg::basename(file_f));
    fibers.load(file_f);
  }
  
  const CImg<unsigned char> line_texture = CImg<>(2,1,1,3).fill(200,255,0,255,0,200).
    RGBtoHSV().resize(256,1,1,3,3).HSVtoRGB();

  // Compute eigen elements
  //------------------------
  std::fprintf(stderr,"\n- Compute eigen elements.");
  CImg<unsigned char> coloredFA(tensors.dimx(),tensors.dimy(),tensors.dimz(),3);
  CImg<> eigen(tensors.dimx(),tensors.dimy(),tensors.dimz(),13);
  CImg<> val,vec;
  float eigmax = 0;
  cimg_mapXYZ(tensors,x,y,z) {
    tensors.get_tensor_at(x,y,z).symmetric_eigen(val,vec);
    eigen(x,y,z,0) = val[0]; eigen(x,y,z,1) = val[1]; eigen(x,y,z,2) = val[2];
    if (val[0]<0) val[0]=0;
    if (val[1]<0) val[1]=0;
    if (val[2]<0) val[2]=0;   
    if (val[0]>eigmax) eigmax = val[0];
    eigen(x,y,z,3) = vec(0,0); eigen(x,y,z,4)  = vec(0,1); eigen(x,y,z,5)  = vec(0,2);
    eigen(x,y,z,6) = vec(1,0); eigen(x,y,z,7)  = vec(1,1); eigen(x,y,z,8)  = vec(1,2);
    eigen(x,y,z,9) = vec(2,0); eigen(x,y,z,10) = vec(2,1); eigen(x,y,z,11) = vec(2,2);
    const float fa = get_FA(val[0],val[1],val[2]);
    eigen(x,y,z,12) = fa;
    const int
      r = (int)cimg::abs(255*fa*vec(0,0)),
      g = (int)cimg::abs(255*fa*vec(0,1)),
      b = (int)cimg::abs(255*fa*vec(0,2));
    coloredFA(x,y,z,0) = (unsigned char)r;
    coloredFA(x,y,z,1) = (unsigned char)g;
    coloredFA(x,y,z,2) = (unsigned char)b;
  }
  tensors.assign();
  CImgStats stats(eigen.get_shared_channel(12),false);
  std::fprintf(stderr,"\n- Maximum diffusivity = %g, Maximum FA = %g",eigmax,stats.max); 
  if (normalize) {
    std::fprintf(stderr,"\n- Normalize tensors.");
    eigen.get_shared_channels(0,2)/=eigmax;
  }
  
  // Init display and begin user interaction
  //-----------------------------------------
  std::fprintf(stderr,"\n- Open user window.");
  CImgDisplay disp(256,256,"DTMRI Viewer",0);
  CImgDisplay disp3d(640,480,"3D Local View",0,3,false,true);
  int s[6];
  unsigned int XYZ[3];
  XYZ[0] = eigen.dimx()/2; XYZ[1] = eigen.dimy()/2; XYZ[2] = eigen.dimz()/2;

  while (!disp.is_closed) {
    coloredFA.feature_selection(s,2,disp,XYZ);
    if (!disp.is_closed) switch (disp.key) {
      
      // Open 3D visualization window
      //-----------------------------
    case cimg::keyA:
    case 0: {
      unsigned char white[1] = { 255 };
      disp3d.display(CImg<unsigned char>(disp3d.dimx(),disp3d.dimy(),1,1,0).draw_text("Please wait...",10,10,white)).show();
      int xm,ym,zm,xM,yM,zM;
      if (!disp.key) { xm=s[0]; ym=s[1]; zm=s[2]; xM=s[3]; yM=s[4]; zM=s[5]; }
      else { xm=ym=zm=0; xM=eigen.dimx()-1; yM=eigen.dimy()-1; zM=eigen.dimy()-1; }
      const CImg<> img = eigen.get_crop(xm,ym,zm,xM,yM,zM);
      CImgList<> points;
      CImgList<unsigned int> primitives;
      CImgList<unsigned char> colors;
      
      // Add ellipsoids to the 3D scene      
      int X = img.dimx()/2, Y = img.dimy()/2, Z = img.dimz()/2;
      { cimg_mapXY(img,x,y) insert_ellipsoid(img.get_vector_at(x,y,Z),(float)x,(float)y,(float)Z,tfact,voxw,voxh,voxd,points,primitives,colors,10,6); }
      { cimg_mapXZ(img,x,z) insert_ellipsoid(img.get_vector_at(x,Y,z),(float)x,(float)Y,(float)z,tfact,voxw,voxh,voxd,points,primitives,colors,10,6); }
      { cimg_mapYZ(img,y,z) insert_ellipsoid(img.get_vector_at(X,y,z),(float)X,(float)y,(float)z,tfact,voxw,voxh,voxd,points,primitives,colors,10,6); }
      
      // Add computed fibers to the 3D scene
      const CImg<> veigen = eigen.get_crop(xm,ym,zm,xM,yM,zM);
      cimglist_map(fibers,l) {
	const CImg<>& fiber = fibers[l];
	if (fiber.dimx()) insert_fiber(fiber,eigen,line_texture,
				       xm,ym,zm,voxw,voxh,voxd,
				       points,primitives,colors);
      }
      
      // Display 3D object
      CImg<unsigned char> visu(disp3d.dimx(),disp3d.dimy(),1,3,0);
      cimg_mapXY(visu,x,y) visu(x,y,2) = y*32/visu.dimy();
      bool stopflag = false;
      while (!disp3d.is_closed && !stopflag) {      
	visu.display_object3d(points,primitives,colors,disp3d,true,4,-1,false,800,0.2f);
	switch (disp3d.key) {
	case cimg::keyF: { // Toggle fullscreen
	  if (disp3d.is_fullscreen) disp3d.resize(800,600);
	  else disp3d.resize(disp3d.screen_dimx(), disp3d.screen_dimy());
	  disp3d.toggle_fullscreen();
	} break;
	case cimg::keyM: { // Create movie
	  std::fprintf(stderr,"\n- Movie mode.\n");
	  const unsigned int N = 256;
	  CImg<> pts = points.get_append('x');
	  CImgList<> cpoints(points);
	  CImg<> x = pts.get_shared_line(0), y = pts.get_shared_line(1), z = pts.get_shared_line(2);
	  const CImgStats sx(x,false), sy(y,false), sz(z,false);
	  const float
	    xm = (float)sx.min, xM = (float)sx.max,
	    ym = (float)sy.min, yM = (float)sy.max,
	    zm = (float)sz.min, zM = (float)sz.max,
	    ratio = 2.0f*cimg::min(visu.dimx(),visu.dimy())/(3.0f*cimg::max(xM-xm,yM-ym,zM-zm)),
	    dx = 0.5f*(xM+xm), dy = 0.5f*(yM+ym), dz = 0.5f*(zM+zm);
	  cimglist_map(points,l) {
	    cpoints(l,0) = (float)((points(l,0)-dx)*ratio);
	    cpoints(l,1) = (float)((points(l,1)-dy)*ratio);
	    cpoints(l,2) = (float)((points(l,2)-dz)*ratio);
	  }
	  	  
	  for (unsigned int i=0; i<N; i++) {
	    std::fprintf(stderr,"\r- Frame %u/%u.",i,N);
	    const float alpha = (float)(i*2*cimg::PI/N);
	    const CImg<> rot = CImg<>::rotation_matrix(0,1,0,alpha)*CImg<>::rotation_matrix(1,0,0,1.30f);
	    CImgList<> rotated(cpoints);
	    cimglist_map(rotated,l) rotated[l] = rot*cpoints[l];
	    visu.fill(0).draw_object3d(visu.dimx()/2.0f,visu.dimy()/2.0f,-500.0f,rotated,primitives,colors,
				       4,false,800.0f,visu.dimx()/2.0f,visu.dimy()/2.0f,-800.0f,0.2f).display(disp3d);
	    visu.save("frame.png",i);
	  }
	  visu.fill(0);
	} break;
	default: stopflag = true;
	}
      }
    } break;
    
    // Compute region statistics
    //---------------------------
    case cimg::keyR: {
      std::fprintf(stderr,"\n- Statistics computation. Select region."); std::fflush(stderr);
      coloredFA.feature_selection(s,2,disp,XYZ);
      int xm,ym,zm,xM,yM,zM;
      if (!disp.key) { xm=s[0]; ym=s[1]; zm=s[2]; xM=s[3]; yM=s[4]; zM=s[5]; }
      else { xm=ym=zm=0; xM=eigen.dimx()-1; yM=eigen.dimy()-1; zM=eigen.dimy()-1; }
      const CImg<> img = eigen.get_crop(xm,ym,zm,xM,yM,zM);
      CImgStats stats(eigen.get_shared_channel(0),false), stats2(eigen.get_shared_channel(12),false);
      std::fprintf(stderr,"\n- Mean diffusivity = %g, Mean FA = %g\n",stats.mean,stats2.mean);		   
    } break;

    // Track fiber bundle (single region)
    //----------------------------------
    case cimg::keyF: {
      std::fprintf(stderr,"\n- Tracking mode (single region). Select starting region."); std::fflush(stderr);
      coloredFA.feature_selection(s,2,disp,XYZ);
      const unsigned int N = fibers.size;
      for (int z=s[2]; z<=s[5]; z++)
	for (int y=s[1]; y<=s[4]; y++)
	  for (int x=s[0]; x<=s[3]; x++) {
	    const CImg<> fiber = get_fibertrack(eigen,(float)x,(float)y,(float)z,lmax,dl,famin,cmin);
	    if (fiber.width>lmin) {
	      std::fprintf(stderr,"Fiber %u : Starting from (%d,%d,%d)\n",fibers.size,x,y,z);			   
	      fibers.insert(fiber);
	    }
	  }
      std::fprintf(stderr," %u fiber(s) added (total %u).",fibers.size-N,fibers.size);
    } break;

    // Track fiber bundle (double regions)
    //------------------------------------
    case cimg::keyG: {
      std::fprintf(stderr,"\n- Tracking mode (double region). Select starting region."); std::fflush(stderr);
      int ns[6];
      coloredFA.feature_selection(s,2,disp,XYZ);
      std::fprintf(stderr," Select ending region."); std::fflush(stderr);
      coloredFA.feature_selection(ns,2,disp,XYZ);
      const unsigned int N = fibers.size;

      // Track from start to end
      for (int z=s[2]; z<=s[5]; z++)
	for (int y=s[1]; y<=s[4]; y++)
	  for (int x=s[0]; x<=s[3]; x++) {
	    const CImg<> fiber = get_fibertrack(eigen,(float)x,(float)y,(float)z,lmax,dl,famin,cmin);
	    if (fiber.width>lmin) {
	      bool valid_fiber = false;
	      cimg_mapX(fiber,k) {
		const int fx = (int)fiber(k,0), fy = (int)fiber(k,1), fz = (int)fiber(k,2);
		if (fx>=ns[0] && fx<=ns[3] &&
		    fy>=ns[1] && fy<=ns[4] &&
		    fz>=ns[2] && fz<=ns[5]) valid_fiber = true;
	      }
	      if (valid_fiber) fibers.insert(fiber);
	    }
	  }
      
      // Track from end to start
      { for (int z=ns[2]; z<=ns[5]; z++)
	for (int y=ns[1]; y<=ns[4]; y++)
	  for (int x=ns[0]; x<=ns[3]; x++) {
	    const CImg<> fiber = get_fibertrack(eigen,(float)x,(float)y,(float)z,lmax,dl,famin,cmin);
	    if (fiber.width>lmin) {
	      bool valid_fiber = false;
	      cimg_mapX(fiber,k) {
		const int fx = (int)fiber(k,0), fy = (int)fiber(k,1), fz = (int)fiber(k,2);
		if (fx>=s[0] && fx<=s[3] &&
		    fy>=s[1] && fy<=s[4] &&
		    fz>=s[2] && fz<=s[5]) valid_fiber = true;
	      }
	      if (valid_fiber) fibers.insert(fiber);
	    }
	  }}
      
      std::fprintf(stderr," %u fiber(s) added (total %u).",fibers.size-N,fibers.size);
    } break;
    
    // Clear fiber bundle
    //-------------------
    case cimg::keyC: {
      std::fprintf(stderr,"\n- Fibers removed.");
      fibers.assign();
    } break;
    
    // Save fibers
    //-------------
    case cimg::keyS: {
      fibers.save("fibers.cimg");
      std::fprintf(stderr,"\n- Fibers saved.");
    } break;

    }
  }

  std::fprintf(stderr,"\n- Exit.\n\n\n");
  return 0;
}
